# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/core/protos/msgbus_kv.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fydb/core/protos/msgbus_kv.proto\x12\rNKikimrClient\"\xd7\x12\n\x10TKeyValueRequest\x12\x10\n\x08TabletId\x18\x01 \x01(\x04\x12\x12\n\nGeneration\x18\x02 \x01(\x04\x12\x0e\n\x06\x43ookie\x18\x03 \x01(\x04\x12G\n\x0e\x43mdDeleteRange\x18\x04 \x03(\x0b\x32/.NKikimrClient.TKeyValueRequest.TCmdDeleteRange\x12W\n\x16\x43mdIncrementGeneration\x18\x05 \x01(\x0b\x32\x37.NKikimrClient.TKeyValueRequest.TCmdIncrementGeneration\x12\x39\n\x07\x43mdRead\x18\x06 \x03(\x0b\x32(.NKikimrClient.TKeyValueRequest.TCmdRead\x12\x43\n\x0c\x43mdReadRange\x18\x07 \x03(\x0b\x32-.NKikimrClient.TKeyValueRequest.TCmdReadRange\x12;\n\x08\x43mdWrite\x18\x08 \x03(\x0b\x32).NKikimrClient.TKeyValueRequest.TCmdWrite\x12=\n\tCmdRename\x18\t \x03(\x0b\x32*.NKikimrClient.TKeyValueRequest.TCmdRename\x12\x43\n\x0c\x43mdCopyRange\x18\x0b \x03(\x0b\x32-.NKikimrClient.TKeyValueRequest.TCmdCopyRange\x12=\n\tCmdConcat\x18\x0c \x03(\x0b\x32*.NKikimrClient.TKeyValueRequest.TCmdConcat\x12\x43\n\x0c\x43mdGetStatus\x18\r \x03(\x0b\x32-.NKikimrClient.TKeyValueRequest.TCmdGetStatus\x12O\n\x12\x43mdTrimLeakedBlobs\x18\x0e \x01(\x0b\x32\x33.NKikimrClient.TKeyValueRequest.TCmdTrimLeakedBlobs\x12\x61\n\x1b\x43mdSetExecutorFastLogPolicy\x18\x0f \x01(\x0b\x32<.NKikimrClient.TKeyValueRequest.TCmdSetExecutorFastLogPolicy\x12\x19\n\x11\x44\x65\x61\x64lineInstantMs\x18\n \x01(\x04\x1aM\n\tTKeyRange\x12\x0c\n\x04\x46rom\x18\x01 \x01(\x0c\x12\x13\n\x0bIncludeFrom\x18\x02 \x01(\x08\x12\n\n\x02To\x18\x03 \x01(\x0c\x12\x11\n\tIncludeTo\x18\x04 \x01(\x08\x1aK\n\x0fTCmdDeleteRange\x12\x38\n\x05Range\x18\x01 \x01(\x0b\x32).NKikimrClient.TKeyValueRequest.TKeyRange\x1a\x19\n\x17TCmdIncrementGeneration\x1ar\n\x08TCmdRead\x12\x0b\n\x03Key\x18\x01 \x01(\x0c\x12\x0e\n\x06Offset\x18\x02 \x01(\x04\x12\x0c\n\x04Size\x18\x03 \x01(\x04\x12;\n\x08Priority\x18\x04 \x01(\x0e\x32).NKikimrClient.TKeyValueRequest.EPriority\x1a\xaf\x01\n\rTCmdReadRange\x12\x38\n\x05Range\x18\x01 \x01(\x0b\x32).NKikimrClient.TKeyValueRequest.TKeyRange\x12\x13\n\x0bIncludeData\x18\x02 \x01(\x08\x12\x12\n\nLimitBytes\x18\x03 \x01(\x04\x12;\n\x08Priority\x18\x04 \x01(\x0e\x32).NKikimrClient.TKeyValueRequest.EPriority\x1a\xd6\x02\n\tTCmdWrite\x12\x0b\n\x03Key\x18\x01 \x01(\x0c\x12\r\n\x05Value\x18\x02 \x01(\x0c\x12G\n\x0eStorageChannel\x18\x03 \x01(\x0e\x32/.NKikimrClient.TKeyValueRequest.EStorageChannel\x12;\n\x08Priority\x18\x04 \x01(\x0e\x32).NKikimrClient.TKeyValueRequest.EPriority\x12\x12\n\nKeyToCache\x18\x05 \x01(\x0c\x12G\n\x06Tactic\x18\x06 \x01(\x0e\x32\'.NKikimrClient.TKeyValueRequest.ETactic:\x0eMAX_THROUGHPUT\x12J\n\x11\x41utoselectChannel\x18\x07 \x03(\x0e\x32/.NKikimrClient.TKeyValueRequest.EStorageChannel\x1a,\n\nTCmdRename\x12\x0e\n\x06OldKey\x18\x01 \x01(\x0c\x12\x0e\n\x06NewKey\x18\x02 \x01(\x0c\x1av\n\rTCmdCopyRange\x12\x38\n\x05Range\x18\x01 \x01(\x0b\x32).NKikimrClient.TKeyValueRequest.TKeyRange\x12\x13\n\x0bPrefixToAdd\x18\x02 \x01(\x0c\x12\x16\n\x0ePrefixToRemove\x18\x03 \x01(\x0c\x1a\x46\n\nTCmdConcat\x12\x11\n\tInputKeys\x18\x01 \x03(\x0c\x12\x11\n\tOutputKey\x18\x02 \x01(\x0c\x12\x12\n\nKeepInputs\x18\x03 \x01(\x08\x1aX\n\rTCmdGetStatus\x12G\n\x0eStorageChannel\x18\x01 \x01(\x0e\x32/.NKikimrClient.TKeyValueRequest.EStorageChannel\x1a-\n\x13TCmdTrimLeakedBlobs\x12\x16\n\x0eMaxItemsToTrim\x18\x01 \x01(\r\x1a\x38\n\x1cTCmdSetExecutorFastLogPolicy\x12\x18\n\tIsAllowed\x18\x01 \x01(\x08:\x05\x66\x61lse\"\x94\x01\n\x0f\x45StorageChannel\x12\x08\n\x04MAIN\x10\x00\x12\t\n\x05\x45XTRA\x10\x01\x12\n\n\x06\x45XTRA2\x10\x02\x12\n\n\x06\x45XTRA3\x10\x03\x12\n\n\x06\x45XTRA4\x10\x04\x12\n\n\x06\x45XTRA5\x10\x05\x12\n\n\x06\x45XTRA6\x10\x06\x12\n\n\x06\x45XTRA7\x10\x07\x12\n\n\x06\x45XTRA8\x10\x08\x12\n\n\x06\x45XTRA9\x10\t\x12\x0c\n\x06INLINE\x10\xff\xff\x03\")\n\tEPriority\x12\x0c\n\x08REALTIME\x10\x00\x12\x0e\n\nBACKGROUND\x10\x01\".\n\x07\x45Tactic\x12\x12\n\x0eMAX_THROUGHPUT\x10\x00\x12\x0f\n\x0bMIN_LATENCY\x10\x01\"\xaf\x0e\n\x11TKeyValueResponse\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12\x16\n\rProtobufError\x18\x90\x03 \x01(\x0c\x12\x0e\n\x06\x43ookie\x18\x02 \x01(\x04\x12N\n\x11\x44\x65leteRangeResult\x18\x03 \x03(\x0b\x32\x33.NKikimrClient.TKeyValueResponse.TDeleteRangeResult\x12^\n\x19IncrementGenerationResult\x18\x04 \x01(\x0b\x32;.NKikimrClient.TKeyValueResponse.TIncrementGenerationResult\x12@\n\nReadResult\x18\x05 \x03(\x0b\x32,.NKikimrClient.TKeyValueResponse.TReadResult\x12J\n\x0fReadRangeResult\x18\x06 \x03(\x0b\x32\x31.NKikimrClient.TKeyValueResponse.TReadRangeResult\x12\x42\n\x0bWriteResult\x18\x07 \x03(\x0b\x32-.NKikimrClient.TKeyValueResponse.TWriteResult\x12\x44\n\x0cRenameResult\x18\x08 \x03(\x0b\x32..NKikimrClient.TKeyValueResponse.TRenameResult\x12J\n\x0f\x43opyRangeResult\x18\n \x03(\x0b\x32\x31.NKikimrClient.TKeyValueResponse.TCopyRangeResult\x12\x44\n\x0c\x43oncatResult\x18\x0b \x03(\x0b\x32..NKikimrClient.TKeyValueResponse.TConcatResult\x12J\n\x0fGetStatusResult\x18\x0c \x03(\x0b\x32\x31.NKikimrClient.TKeyValueResponse.TGetStatusResult\x12V\n\x15TrimLeakedBlobsResult\x18\r \x01(\x0b\x32\x37.NKikimrClient.TKeyValueResponse.TTrimLeakedBlobsResult\x12h\n\x1eSetExecutorFastLogPolicyResult\x18\x0e \x01(\x0b\x32@.NKikimrClient.TKeyValueResponse.TSetExecutorFastLogPolicyResult\x12\x13\n\x0b\x45rrorReason\x18\t \x01(\t\x1a\xb1\x01\n\rTKeyValuePair\x12\x0b\n\x03Key\x18\x01 \x01(\x0c\x12\r\n\x05Value\x18\x02 \x01(\x0c\x12\x11\n\tValueSize\x18\x03 \x01(\r\x12\x18\n\x10\x43reationUnixTime\x18\x04 \x01(\x04\x12G\n\x0eStorageChannel\x18\x05 \x01(\x0e\x32/.NKikimrClient.TKeyValueRequest.EStorageChannel\x12\x0e\n\x06Status\x18\x06 \x01(\r\x1a$\n\x12TDeleteRangeResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x1a@\n\x1aTIncrementGenerationResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12\x12\n\nGeneration\x18\x02 \x01(\x04\x1a=\n\x0bTReadResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12\r\n\x05Value\x18\x02 \x01(\x0c\x12\x0f\n\x07Message\x18\x03 \x01(\t\x1a`\n\x10TReadRangeResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12<\n\x04Pair\x18\x02 \x03(\x0b\x32..NKikimrClient.TKeyValueResponse.TKeyValuePair\x1a\x33\n\x0cTWriteResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12\x13\n\x0bStatusFlags\x18\x02 \x01(\r\x1a\x1f\n\rTRenameResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x1a\"\n\x10TCopyRangeResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x1a\x1f\n\rTConcatResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x1a\x80\x01\n\x10TGetStatusResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12G\n\x0eStorageChannel\x18\x02 \x01(\x0e\x32/.NKikimrClient.TKeyValueRequest.EStorageChannel\x12\x13\n\x0bStatusFlags\x18\x03 \x01(\r\x1aW\n\x16TTrimLeakedBlobsResult\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12\x17\n\x0fNumItemsTrimmed\x18\x02 \x01(\r\x12\x14\n\x0cNumItemsLeft\x18\x03 \x01(\r\x1a\x31\n\x1fTSetExecutorFastLogPolicyResult\x12\x0e\n\x06Status\x18\x01 \x01(\rB\x18\n\x16ru.yandex.kikimr.proto')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.core.protos.msgbus_kv_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026ru.yandex.kikimr.proto'
  _TKEYVALUEREQUEST._serialized_start=51
  _TKEYVALUEREQUEST._serialized_end=2442
  _TKEYVALUEREQUEST_TKEYRANGE._serialized_start=947
  _TKEYVALUEREQUEST_TKEYRANGE._serialized_end=1024
  _TKEYVALUEREQUEST_TCMDDELETERANGE._serialized_start=1026
  _TKEYVALUEREQUEST_TCMDDELETERANGE._serialized_end=1101
  _TKEYVALUEREQUEST_TCMDINCREMENTGENERATION._serialized_start=1103
  _TKEYVALUEREQUEST_TCMDINCREMENTGENERATION._serialized_end=1128
  _TKEYVALUEREQUEST_TCMDREAD._serialized_start=1130
  _TKEYVALUEREQUEST_TCMDREAD._serialized_end=1244
  _TKEYVALUEREQUEST_TCMDREADRANGE._serialized_start=1247
  _TKEYVALUEREQUEST_TCMDREADRANGE._serialized_end=1422
  _TKEYVALUEREQUEST_TCMDWRITE._serialized_start=1425
  _TKEYVALUEREQUEST_TCMDWRITE._serialized_end=1767
  _TKEYVALUEREQUEST_TCMDRENAME._serialized_start=1769
  _TKEYVALUEREQUEST_TCMDRENAME._serialized_end=1813
  _TKEYVALUEREQUEST_TCMDCOPYRANGE._serialized_start=1815
  _TKEYVALUEREQUEST_TCMDCOPYRANGE._serialized_end=1933
  _TKEYVALUEREQUEST_TCMDCONCAT._serialized_start=1935
  _TKEYVALUEREQUEST_TCMDCONCAT._serialized_end=2005
  _TKEYVALUEREQUEST_TCMDGETSTATUS._serialized_start=2007
  _TKEYVALUEREQUEST_TCMDGETSTATUS._serialized_end=2095
  _TKEYVALUEREQUEST_TCMDTRIMLEAKEDBLOBS._serialized_start=2097
  _TKEYVALUEREQUEST_TCMDTRIMLEAKEDBLOBS._serialized_end=2142
  _TKEYVALUEREQUEST_TCMDSETEXECUTORFASTLOGPOLICY._serialized_start=2144
  _TKEYVALUEREQUEST_TCMDSETEXECUTORFASTLOGPOLICY._serialized_end=2200
  _TKEYVALUEREQUEST_ESTORAGECHANNEL._serialized_start=2203
  _TKEYVALUEREQUEST_ESTORAGECHANNEL._serialized_end=2351
  _TKEYVALUEREQUEST_EPRIORITY._serialized_start=2353
  _TKEYVALUEREQUEST_EPRIORITY._serialized_end=2394
  _TKEYVALUEREQUEST_ETACTIC._serialized_start=2396
  _TKEYVALUEREQUEST_ETACTIC._serialized_end=2442
  _TKEYVALUERESPONSE._serialized_start=2445
  _TKEYVALUERESPONSE._serialized_end=4284
  _TKEYVALUERESPONSE_TKEYVALUEPAIR._serialized_start=3416
  _TKEYVALUERESPONSE_TKEYVALUEPAIR._serialized_end=3593
  _TKEYVALUERESPONSE_TDELETERANGERESULT._serialized_start=3595
  _TKEYVALUERESPONSE_TDELETERANGERESULT._serialized_end=3631
  _TKEYVALUERESPONSE_TINCREMENTGENERATIONRESULT._serialized_start=3633
  _TKEYVALUERESPONSE_TINCREMENTGENERATIONRESULT._serialized_end=3697
  _TKEYVALUERESPONSE_TREADRESULT._serialized_start=3699
  _TKEYVALUERESPONSE_TREADRESULT._serialized_end=3760
  _TKEYVALUERESPONSE_TREADRANGERESULT._serialized_start=3762
  _TKEYVALUERESPONSE_TREADRANGERESULT._serialized_end=3858
  _TKEYVALUERESPONSE_TWRITERESULT._serialized_start=3860
  _TKEYVALUERESPONSE_TWRITERESULT._serialized_end=3911
  _TKEYVALUERESPONSE_TRENAMERESULT._serialized_start=3913
  _TKEYVALUERESPONSE_TRENAMERESULT._serialized_end=3944
  _TKEYVALUERESPONSE_TCOPYRANGERESULT._serialized_start=3946
  _TKEYVALUERESPONSE_TCOPYRANGERESULT._serialized_end=3980
  _TKEYVALUERESPONSE_TCONCATRESULT._serialized_start=3982
  _TKEYVALUERESPONSE_TCONCATRESULT._serialized_end=4013
  _TKEYVALUERESPONSE_TGETSTATUSRESULT._serialized_start=4016
  _TKEYVALUERESPONSE_TGETSTATUSRESULT._serialized_end=4144
  _TKEYVALUERESPONSE_TTRIMLEAKEDBLOBSRESULT._serialized_start=4146
  _TKEYVALUERESPONSE_TTRIMLEAKEDBLOBSRESULT._serialized_end=4233
  _TKEYVALUERESPONSE_TSETEXECUTORFASTLOGPOLICYRESULT._serialized_start=4235
  _TKEYVALUERESPONSE_TSETEXECUTORFASTLOGPOLICYRESULT._serialized_end=4284
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/library/yql/dq/proto/dq_transport.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+ydb/library/yql/dq/proto/dq_transport.proto\x12\rNYql.NDqProto\"<\n\x05TData\x12\x18\n\x10TransportVersion\x18\x01 \x01(\r\x12\x0b\n\x03Raw\x18\x02 \x01(\x0c\x12\x0c\n\x04Rows\x18\x03 \x01(\r*\xa1\x01\n\x15\x45\x44\x61taTransportVersion\x12&\n\"DATA_TRANSPORT_VERSION_UNSPECIFIED\x10\x00\x12\x1c\n\x17\x44\x41TA_TRANSPORT_YSON_1_0\x10\x90N\x12\"\n\x1c\x44\x41TA_TRANSPORT_UV_PICKLE_1_0\x10\xa0\x9c\x01\x12\x1e\n\x18\x44\x41TA_TRANSPORT_ARROW_1_0\x10\xb0\xea\x01\x42\x03\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.library.yql.dq.proto.dq_transport_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _EDATATRANSPORTVERSION._serialized_start=125
  _EDATATRANSPORTVERSION._serialized_end=286
  _TDATA._serialized_start=62
  _TDATA._serialized_end=122
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/library/yql/public/issue/protos/issue_message.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7ydb/library/yql/public/issue/protos/issue_message.proto\x12\x12NYql.NIssue.NProto\"\xac\x02\n\x0cIssueMessage\x12;\n\x08position\x18\x01 \x01(\x0b\x32).NYql.NIssue.NProto.IssueMessage.Position\x12\x0f\n\x07message\x18\x02 \x01(\t\x12?\n\x0c\x65nd_position\x18\x03 \x01(\x0b\x32).NYql.NIssue.NProto.IssueMessage.Position\x12\x12\n\nissue_code\x18\x04 \x01(\r\x12\x10\n\x08severity\x18\x05 \x01(\r\x12\x30\n\x06issues\x18\x06 \x03(\x0b\x32 .NYql.NIssue.NProto.IssueMessage\x1a\x35\n\x08Position\x12\x0b\n\x03row\x18\x01 \x01(\r\x12\x0e\n\x06\x63olumn\x18\x02 \x01(\r\x12\x0c\n\x04\x66ile\x18\x03 \x01(\tB\x1c\n\x1a\x63om.yandex.yql.issue.proto')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.library.yql.public.issue.protos.issue_message_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.yandex.yql.issue.proto'
  _ISSUEMESSAGE._serialized_start=80
  _ISSUEMESSAGE._serialized_end=380
  _ISSUEMESSAGE_POSITION._serialized_start=327
  _ISSUEMESSAGE_POSITION._serialized_end=380
# @@protoc_insertion_point(module_scope)



import ydb.apps.dstool.lib.common as common
import ydb.apps.dstool.lib.table as table
import json
import sys

description = 'Get blob information from group'


def add_options(p):
    p.add_argument('--group-id', type=int, required=True, help='Group id')
    p.add_argument('--blob-id', type=str, required=True, help='Blob id, e.g. [72075186224037917:2020:4:2:72707:6337498:0]')
    table.TableOutput([]).add_options(p)


def process_vslot(vslot, node_mon_map, blob_id):
    page = 'vdisk/json/getblob'
    params = {
        'node_id': vslot.VSlotId.NodeId,
        'pdisk_id': vslot.VSlotId.PDiskId,
        'vslot_id': vslot.VSlotId.VSlotId,
        'from': blob_id,
        'to': blob_id,
        'internals': 'yes'
    }
    host = node_mon_map[vslot.VSlotId.NodeId]
    data = common.fetch(page, params, host, 'raw').decode('utf-8')
    data = json.loads(data)
    res = []
    if 'logoblobs' in data:
        for blob in data['logoblobs']:
            row = {}
            row['BlobId'] = blob['id']
            row['Status'] = blob['status']
            row['Ingress'] = blob['ingress']
            res.append(row)

    return res


def do(args):
    try:
        columns = [
            'BlobId',
            'Status',
            'Ingress',
        ]

        table_output = table.TableOutput(columns, default_visible_columns=columns)

        base_config = common.fetch_base_config()
        group_map = common.build_group_map(base_config)
        if args.group_id not in group_map:
            raise Exception('Unknown group with id %u' % args.group_id)

        group = group_map[args.group_id]
        vslot_map = common.build_vslot_map(base_config)
        node_mon_map = common.fetch_node_mon_map()

        rows = []
        for vslot in common.vslots_of_group(group, vslot_map):
            res = process_vslot(vslot, node_mon_map, args.blob_id)
            rows.extend(res)

        if rows:
            table_output.dump(rows, args)
    except Exception as e:
        common.print_status(args, success=False, error_reason=e)
        sys.exit(1)



import ydb.apps.dstool.lib.common as common
import time
import sys
import random
from collections import defaultdict, Counter

description = 'Move vdisks out from overpopulated pdisks.'


def add_options(p):
    p.add_argument('--max-replicating-pdisks', type=int, help='Limit number of maximum replicating PDisks in the cluster')
    p.add_argument('--only-from-overpopulated-pdisks', action='store_true', help='Move vdisks out only from pdisks with over expected slot count')
    common.add_basic_format_options(p)


def do(args):
    while True:
        common.flush_cache()

        base_config = common.fetch_base_config()
        node_mon_map = common.fetch_node_mon_map({vslot.VSlotId.NodeId for vslot in base_config.VSlot})
        vslot_map = common.build_vslot_map(base_config)
        pdisk_map = common.build_pdisk_map(base_config)
        pdisk_usage = common.build_pdisk_usage_map(base_config)

        vdisks_groups_count_map = defaultdict(int)
        for group in base_config.Group:
            num = sum(vslot.Status == 'READY' for vslot in common.vslots_of_group(group, vslot_map)) - len(group.VSlotId)
            vdisks_groups_count_map[num] += 1

        if any(k < -1 for k in vdisks_groups_count_map.keys()):
            common.print_if_not_quiet(args, 'There are groups with more than one non READY vslot, waiting...', sys.stdout)
            common.print_if_verbose(args, f'Number of non READY vdisks -> number of groups: {sorted(vdisks_groups_count_map.items())}', file=sys.stdout)
            time.sleep(15)
            continue

        if args.max_replicating_pdisks is not None:
            replicating_pdisks = set()
            for vslot in base_config.VSlot:
                if vslot.Status != 'READY' and vslot.Status != 'ERROR':
                    replicating_pdisks.add(common.get_pdisk_id(vslot.VSlotId))

            if len(replicating_pdisks) > args.max_replicating_pdisks:
                common.print_if_not_quiet(args, 'Waiting for %d pdisks to finish replication...' % (len(replicating_pdisks) - args.max_replicating_pdisks), sys.stdout)
                common.print_if_verbose(args, 'Replicating pdisks: ' + ', '.join('[%d:%d]' % x for x in sorted(replicating_pdisks)), file=sys.stdout)
                time.sleep(15)
                continue

        all_groups = common.select_groups(base_config)
        healthy_groups = common.filter_healthy_groups(all_groups, node_mon_map, base_config, vslot_map)
        unhealthy_groups = all_groups - healthy_groups
        if unhealthy_groups:
            common.print_if_verbose(args, 'Skipping vdisks from unhealthy groups: %s' % (unhealthy_groups), file=sys.stdout)

        healty_vslots = [
            vslot
            for vslot in base_config.VSlot
            if vslot.GroupId in healthy_groups
        ]

        overpopulated_pdisks = set()
        for pdisk_id in pdisk_map.keys():
            if pdisk_map[pdisk_id].ExpectedSlotCount and pdisk_usage[pdisk_id] > pdisk_map[pdisk_id].ExpectedSlotCount:
                overpopulated_pdisks.add(pdisk_id)

        if not overpopulated_pdisks:
            common.print_if_not_quiet(args, 'No overpopulated pdisks found', sys.stdout)
            if args.only_from_overpopulated_pdisks:
                common.print_status(args, success=True, error_reason='')
                break

        healty_vslots_from_overpopulated_pdisks = []
        for vslot in base_config.VSlot:
            pdisk_id = common.get_pdisk_id(vslot.VSlotId)
            if pdisk_id not in overpopulated_pdisks:
                continue
            if vslot.GroupId not in healthy_groups:
                continue

            healty_vslots_from_overpopulated_pdisks.append(vslot)

        candidate_vslots = []
        if healty_vslots_from_overpopulated_pdisks:
            common.print_if_not_quiet(args, f'Found {len(healty_vslots_from_overpopulated_pdisks)} vdisks from overpopulated pdisks', sys.stdout)
            candidate_vslots = healty_vslots_from_overpopulated_pdisks
        elif healty_vslots and not args.only_from_overpopulated_pdisks:
            common.print_if_not_quiet(args, f'Found {len(healty_vslots)} vdisks suitable for relocation', sys.stdout)
            candidate_vslots = healty_vslots
        else:  # candidate_vslots is empty
            common.print_if_not_quiet(args, 'No vdisks suitable for relocation found, waiting..', sys.stdout)
            time.sleep(10)
            continue

        histo = Counter(pdisk_usage.values())
        common.print_if_verbose(args, 'Number of used slots -> number pdisks: ' + ' '.join('%d=>%d' % (k, histo[k]) for k in sorted(histo)), file=sys.stdout)

        def do_reassign(vslot, try_blocking):
            pdisk_id = common.get_pdisk_id(vslot.VSlotId)
            vslot_id = common.get_vslot_id(vslot.VSlotId)

            common.print_if_verbose(args, 'Checking to relocate vdisk from vslot %s on pdisk %s with slot usage %d' % (vslot_id, pdisk_id, pdisk_usage[pdisk_id]), file=sys.stdout)

            current_usage = pdisk_usage[pdisk_id]
            if not healty_vslots_from_overpopulated_pdisks:
                for i in range(0, current_usage - 1):
                    if histo[i]:
                        break
                else:
                    return False

            def add_update_drive_status(request, pdisk, status):
                cmd = request.Command.add().UpdateDriveStatus
                cmd.HostKey.NodeId = pdisk.NodeId
                cmd.PDiskId = pdisk.PDiskId
                cmd.Status = status

            def add_reassign_cmd(request, vslot):
                cmd = request.Command.add().ReassignGroupDisk
                cmd.GroupId = vslot.GroupId
                cmd.GroupGeneration = vslot.GroupGeneration
                cmd.FailRealmIdx = vslot.FailRealmIdx
                cmd.FailDomainIdx = vslot.FailDomainIdx
                cmd.VDiskIdx = vslot.VDiskIdx

            request = common.kikimr_bsconfig.TConfigRequest(Rollback=True)
            index = len(request.Command)
            add_reassign_cmd(request, vslot)
            response = common.invoke_bsc_request(request)
            if len(response.Status) != 1 or not response.Status[0].Success:
                return False
            item = response.Status[index].ReassignedItem[0]
            pdisk_from = item.From.NodeId, item.From.PDiskId
            pdisk_to = item.To.NodeId, item.To.PDiskId
            if pdisk_usage[pdisk_to] + 1 > pdisk_usage[pdisk_from] - 1:
                assert not healty_vslots_from_overpopulated_pdisks
                if not try_blocking:
                    return False
                request = common.kikimr_bsconfig.TConfigRequest(Rollback=True)
                inactive = []
                for pdisk in base_config.PDisk:
                    if pdisk_usage[common.get_pdisk_id(pdisk)] + 1 > pdisk_usage[pdisk_id] - 1:
                        add_update_drive_status(request, pdisk, common.kikimr_bsconfig.EDriveStatus.INACTIVE)
                        inactive.append(pdisk)
                index = len(request.Command)
                add_reassign_cmd(request, vslot)
                for pdisk in inactive:
                    add_update_drive_status(request, pdisk, pdisk.DriveStatus)
                response = common.invoke_bsc_request(request)
                if len(response.Status) != 1 or not response.Status[index].Success:
                    return False

            request.Rollback = args.dry_run
            response = common.invoke_bsc_request(request)

            if response.Status[index].Success:
                from_pdisk_id = common.get_pdisk_id(response.Status[index].ReassignedItem[0].From)
                to_pdisk_id = common.get_pdisk_id(response.Status[index].ReassignedItem[0].To)
                common.print_if_not_quiet(
                    args,
                    'Relocated vdisk from pdisk [%d:%d] to pdisk [%d:%d] with slot usages (%d -> %d)' % (*from_pdisk_id, *to_pdisk_id, pdisk_usage[from_pdisk_id], pdisk_usage[to_pdisk_id]),
                    file=sys.stdout)

            if not common.is_successful_bsc_response(response):
                common.print_request_result(args, request, response)
                sys.exit(1)

            return True
        # end of do_reassign()

        vslots_by_pdisk_slot_usage = defaultdict(list)
        for vslot in candidate_vslots:
            pdisk_id = common.get_pdisk_id(vslot.VSlotId)
            pdisk_slot_usage = pdisk_usage[pdisk_id]
            vslots_by_pdisk_slot_usage[pdisk_slot_usage].append(vslot)

        # check vslots from pdisks with the highest slot usage first
        for pdisk_slot_usage, vslots in sorted(vslots_by_pdisk_slot_usage.items(), reverse=True):
            random.shuffle(vslots)
            for vslot in vslots:
                if do_reassign(vslot, False):
                    break
            else:
                for vslot in vslots:
                    if do_reassign(vslot, True):
                        break
                else:
                    continue
            break
        else:
            common.print_status(args, success=True, error_reason='')
            break



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/core/protos/load_test.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb.core.protos import blobstorage_pb2 as ydb_dot_core_dot_protos_dot_blobstorage__pb2
from ydb.core.protos import blobstorage_disk_pb2 as ydb_dot_core_dot_protos_dot_blobstorage__disk__pb2
from ydb.core.protos import datashard_load_pb2 as ydb_dot_core_dot_protos_dot_datashard__load__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fydb/core/protos/load_test.proto\x12\x07NKikimr\x1a!ydb/core/protos/blobstorage.proto\x1a&ydb/core/protos/blobstorage_disk.proto\x1a$ydb/core/protos/datashard_load.proto\"\xd8)\n\x12TEvLoadTestRequest\x12\x0e\n\x06\x43ookie\x18\x01 \x01(\x04\x12\x31\n\x04Stop\x18\x03 \x01(\x0b\x32!.NKikimr.TEvLoadTestRequest.TStopH\x00\x12?\n\x0bStorageLoad\x18\x02 \x01(\x0b\x32(.NKikimr.TEvLoadTestRequest.TStorageLoadH\x00\x12\x45\n\x0ePDiskWriteLoad\x18\x04 \x01(\x0b\x32+.NKikimr.TEvLoadTestRequest.TPDiskWriteLoadH\x00\x12;\n\tVDiskLoad\x18\x05 \x01(\x0b\x32&.NKikimr.TEvLoadTestRequest.TVDiskLoadH\x00\x12\x43\n\rPDiskReadLoad\x18\x06 \x01(\x0b\x32*.NKikimr.TEvLoadTestRequest.TPDiskReadLoadH\x00\x12\x41\n\x0cPDiskLogLoad\x18\x07 \x01(\x0b\x32).NKikimr.TEvLoadTestRequest.TPDiskLogLoadH\x00\x12\x41\n\x0cKeyValueLoad\x18\x08 \x01(\x0b\x32).NKikimr.TEvLoadTestRequest.TKeyValueLoadH\x00\x12\x37\n\x07KqpLoad\x18\t \x01(\x0b\x32$.NKikimr.TEvLoadTestRequest.TKqpLoadH\x00\x12=\n\nMemoryLoad\x18\n \x01(\x0b\x32\'.NKikimr.TEvLoadTestRequest.TMemoryLoadH\x00\x12@\n\x08YCSBLoad\x18\x0b \x01(\x0b\x32,.NKikimrDataShardLoad.TEvYCSBTestLoadRequestH\x00\x1a\xdb\x02\n\rTIntervalInfo\x12\x0e\n\x06Weight\x18\x01 \x01(\x01\x12M\n\x07Uniform\x18\x02 \x01(\x0b\x32:.NKikimr.TEvLoadTestRequest.TIntervalInfo.TIntervalUniformH\x00\x12M\n\x07Poisson\x18\x03 \x01(\x0b\x32:.NKikimr.TEvLoadTestRequest.TIntervalInfo.TIntervalPoissonH\x00\x1aN\n\x10TIntervalUniform\x12\r\n\x05MinMs\x18\x01 \x01(\r\x12\r\n\x05MaxMs\x18\x02 \x01(\r\x12\r\n\x05MinUs\x18\x03 \x01(\r\x12\r\n\x05MaxUs\x18\x04 \x01(\r\x1a<\n\x10TIntervalPoisson\x12\x11\n\tFrequency\x18\x01 \x01(\x01\x12\x15\n\rMaxIntervalMs\x18\x02 \x01(\rB\x0e\n\x0c\x44istribution\x1a\x35\n\tTSizeInfo\x12\x0b\n\x03Min\x18\x01 \x01(\r\x12\x0b\n\x03Max\x18\x02 \x01(\r\x12\x0e\n\x06Weight\x18\x03 \x01(\x01\x1a\xc9\t\n\x0cTStorageLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x17\n\x0f\x44urationSeconds\x18\x02 \x01(\r\x12\x1e\n\x0fRequestTracking\x18\x03 \x01(\x08:\x05\x66\x61lse\x12K\n\x07Tablets\x18\x04 \x03(\x0b\x32:.NKikimr.TEvLoadTestRequest.TStorageLoad.TPerTabletProfile\x12\x1b\n\x13ScheduleThresholdUs\x18\x05 \x01(\x04\x12\x1a\n\x12ScheduleRoundingUs\x18\x06 \x01(\x04\x1ay\n\x0cTRequestInfo\x12\x10\n\x08SendTime\x18\x01 \x01(\x02\x12\x0c\n\x04Type\x18\x02 \x01(\x04\x12\x0c\n\x04Size\x18\x03 \x01(\r\x12;\n\x0ePutHandleClass\x18\x04 \x01(\x0e\x32#.NKikimrBlobStorage.EPutHandleClass\x1a\xc0\x01\n\x0bTTabletInfo\x12\x10\n\x08TabletId\x18\x01 \x01(\x04\x12\x0f\n\x07\x43hannel\x18\x02 \x01(\r\x12\x0f\n\x07GroupId\x18\x03 \x01(\r\x12\x12\n\nGeneration\x18\x04 \x01(\r\x12G\n\x08Requests\x18\x05 \x03(\x0b\x32\x35.NKikimr.TEvLoadTestRequest.TStorageLoad.TRequestInfo\x12 \n\x18ScriptedCycleDurationSec\x18\x06 \x01(\x02\x1a\xae\x05\n\x11TPerTabletProfile\x12\x45\n\x07Tablets\x18\x01 \x03(\x0b\x32\x34.NKikimr.TEvLoadTestRequest.TStorageLoad.TTabletInfo\x12\x34\n\x05Sizes\x18\x02 \x03(\x0b\x32%.NKikimr.TEvLoadTestRequest.TSizeInfo\x12\x41\n\x0eWriteIntervals\x18\x03 \x03(\x0b\x32).NKikimr.TEvLoadTestRequest.TIntervalInfo\x12\x1b\n\x13MaxInFlightRequests\x18\x04 \x01(\r\x12\x18\n\x10MaxInFlightBytes\x18\x05 \x01(\r\x12\x41\n\x0e\x46lushIntervals\x18\x06 \x03(\x0b\x32).NKikimr.TEvLoadTestRequest.TIntervalInfo\x12;\n\x0ePutHandleClass\x18\x07 \x01(\x0e\x32#.NKikimrBlobStorage.EPutHandleClass\x12\x0c\n\x04Soft\x18\x08 \x01(\x08\x12\x1f\n\x17MaxInFlightReadRequests\x18\t \x01(\r\x12\x1c\n\x14MaxInFlightReadBytes\x18\n \x01(\r\x12@\n\rReadIntervals\x18\x0b \x03(\x0b\x32).NKikimr.TEvLoadTestRequest.TIntervalInfo\x12\x38\n\tReadSizes\x18\x0c \x03(\x0b\x32%.NKikimr.TEvLoadTestRequest.TSizeInfo\x12\x1c\n\x14MaxTotalBytesWritten\x18\r \x01(\x04\x12;\n\x0eGetHandleClass\x18\x0e \x01(\x0e\x32#.NKikimrBlobStorage.EGetHandleClass\x1a+\n\x05TStop\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x15\n\rRemoveAllTags\x18\x02 \x01(\x08\x1a\xcd\x03\n\x0fTPDiskWriteLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x0f\n\x07PDiskId\x18\x02 \x01(\r\x12\x11\n\tPDiskGuid\x18\x03 \x01(\x04\x12-\n\x07VDiskId\x18\x04 \x01(\x0b\x32\x1c.NKikimrBlobStorage.TVDiskID\x12\x46\n\x06\x43hunks\x18\x05 \x03(\x0b\x32\x36.NKikimr.TEvLoadTestRequest.TPDiskWriteLoad.TChunkInfo\x12\x17\n\x0f\x44urationSeconds\x18\x06 \x01(\r\x12\x16\n\x0eInFlightWrites\x18\x07 \x01(\r\x12\x35\n\x07LogMode\x18\x08 \x01(\x0e\x32$.NKikimr.TEvLoadTestRequest.ELogMode\x12\x18\n\nSequential\x18\t \x01(\x08:\x04true\x12\x15\n\rIntervalMsMin\x18\n \x01(\r\x12\x15\n\rIntervalMsMax\x18\x0b \x01(\r\x12\x14\n\x05Reuse\x18\x0c \x01(\x08:\x05\x66\x61lse\x12\x1f\n\x10IsWardenlessTest\x18\r \x01(\x08:\x05\x66\x61lse\x1a+\n\nTChunkInfo\x12\r\n\x05Slots\x18\x01 \x01(\r\x12\x0e\n\x06Weight\x18\x02 \x01(\r\x1a\xaa\x04\n\nTVDiskLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12-\n\x07VDiskId\x18\x02 \x01(\x0b\x32\x1c.NKikimrBlobStorage.TVDiskID\x12\x31\n\tGroupInfo\x18\x10 \x01(\x0b\x32\x1e.NKikimrBlobStorage.TGroupInfo\x12\x10\n\x08TabletId\x18\x05 \x01(\x04\x12\x0f\n\x07\x43hannel\x18\x06 \x01(\r\x12\x12\n\nGeneration\x18\x07 \x01(\r\x12\x17\n\x0f\x44urationSeconds\x18\x08 \x01(\r\x12\x41\n\x0eWriteIntervals\x18\t \x03(\x0b\x32).NKikimr.TEvLoadTestRequest.TIntervalInfo\x12\x39\n\nWriteSizes\x18\n \x03(\x0b\x32%.NKikimr.TEvLoadTestRequest.TSizeInfo\x12\x17\n\x0fInFlightPutsMax\x18\x0b \x01(\r\x12\x1b\n\x13InFlightPutBytesMax\x18\x0c \x01(\x04\x12;\n\x0ePutHandleClass\x18\r \x01(\x0e\x32#.NKikimrBlobStorage.EPutHandleClass\x12J\n\x17\x42\x61rrierAdvanceIntervals\x18\x0e \x03(\x0b\x32).NKikimr.TEvLoadTestRequest.TIntervalInfo\x12\x14\n\x0cStepDistance\x18\x0f \x01(\rJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05\x1a\xfe\x02\n\x0eTPDiskReadLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x0f\n\x07PDiskId\x18\x02 \x01(\r\x12\x11\n\tPDiskGuid\x18\x03 \x01(\x04\x12-\n\x07VDiskId\x18\x04 \x01(\x0b\x32\x1c.NKikimrBlobStorage.TVDiskID\x12\x45\n\x06\x43hunks\x18\x05 \x03(\x0b\x32\x35.NKikimr.TEvLoadTestRequest.TPDiskReadLoad.TChunkInfo\x12\x17\n\x0f\x44urationSeconds\x18\x06 \x01(\r\x12\x15\n\rInFlightReads\x18\x07 \x01(\r\x12\x19\n\nSequential\x18\t \x01(\x08:\x05\x66\x61lse\x12\x15\n\rIntervalMsMin\x18\n \x01(\r\x12\x15\n\rIntervalMsMax\x18\x0b \x01(\r\x12\x1f\n\x10IsWardenlessTest\x18\r \x01(\x08:\x05\x66\x61lse\x1a+\n\nTChunkInfo\x12\r\n\x05Slots\x18\x01 \x01(\r\x12\x0e\n\x06Weight\x18\x02 \x01(\r\x1a\xad\x03\n\rTPDiskLogLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x0f\n\x07PDiskId\x18\x02 \x01(\r\x12\x11\n\tPDiskGuid\x18\x03 \x01(\x04\x12\x17\n\x0f\x44urationSeconds\x18\x05 \x01(\r\x12H\n\x07Workers\x18\x07 \x03(\x0b\x32\x37.NKikimr.TEvLoadTestRequest.TPDiskLogLoad.TWorkerConfig\x12\x1f\n\x10IsWardenlessTest\x18\x08 \x01(\x08:\x05\x66\x61lse\x1a\xe6\x01\n\rTWorkerConfig\x12-\n\x07VDiskId\x18\x01 \x01(\x0b\x32\x1c.NKikimrBlobStorage.TVDiskID\x12\x13\n\x0bMaxInFlight\x18\x02 \x01(\r\x12\x17\n\x0fSizeIntervalMin\x18\x03 \x01(\r\x12\x17\n\x0fSizeIntervalMax\x18\x04 \x01(\r\x12\x15\n\rBurstInterval\x18\x05 \x01(\x04\x12\x11\n\tBurstSize\x18\x06 \x01(\x04\x12\x17\n\x0fStorageDuration\x18\x07 \x01(\x04\x12\x1c\n\x14MaxTotalBytesWritten\x18\x08 \x01(\x04\x1a\x92\x02\n\rTKeyValueLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x16\n\x0eTargetTabletId\x18\x02 \x01(\x04\x12\x17\n\x0f\x44urationSeconds\x18\x05 \x01(\r\x12H\n\x07Workers\x18\x07 \x03(\x0b\x32\x37.NKikimr.TEvLoadTestRequest.TKeyValueLoad.TWorkerConfig\x1ay\n\rTWorkerConfig\x12\x11\n\tKeyPrefix\x18\x01 \x01(\t\x12\x13\n\x0bMaxInFlight\x18\x02 \x01(\r\x12\x0c\n\x04Size\x18\x0b \x01(\r\x12\x17\n\x08IsInline\x18\t \x01(\x08:\x05\x66\x61lse\x12\x19\n\x0eLoopAtKeyCount\x18\n \x01(\r:\x01\x30\x1a\xbc\x05\n\x08TKqpLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x17\n\x0f\x44urationSeconds\x18\x02 \x01(\r\x12\x16\n\x0eWindowDuration\x18\x03 \x01(\r\x12\x12\n\nWorkingDir\x18\x04 \x01(\t\x12\x15\n\rNumOfSessions\x18\x05 \x01(\r\x12\x18\n\x10IncreaseSessions\x18\x0b \x01(\x08\x12\x1b\n\x13\x44\x65leteTableOnFinish\x18\x06 \x01(\x08\x12\x1e\n\x16UniformPartitionsCount\x18\x07 \x01(\r\x12\x14\n\x0cWorkloadType\x18\x08 \x01(\r\x12\x44\n\x05Stock\x18\t \x01(\x0b\x32\x33.NKikimr.TEvLoadTestRequest.TKqpLoad.TStockWorkloadH\x00\x12>\n\x02Kv\x18\n \x01(\x0b\x32\x30.NKikimr.TEvLoadTestRequest.TKqpLoad.TKvWorkloadH\x00\x1a\x8f\x01\n\x0eTStockWorkload\x12\x19\n\x0cProductCount\x18\x01 \x01(\x04:\x03\x31\x30\x30\x12\x16\n\x08Quantity\x18\x02 \x01(\x04:\x04\x31\x30\x30\x30\x12\x17\n\nOrderCount\x18\x03 \x01(\x04:\x03\x31\x30\x30\x12\x11\n\x05Limit\x18\x04 \x01(\x04:\x02\x31\x30\x12\x1e\n\x10PartitionsByLoad\x18\x05 \x01(\x08:\x04true\x1a\xb5\x01\n\x0bTKvWorkload\x12\x1a\n\x0cInitRowCount\x18\x01 \x01(\x04:\x04\x31\x30\x30\x30\x12\x1e\n\x10PartitionsByLoad\x18\x02 \x01(\x08:\x04true\x12)\n\x0bMaxFirstKey\x18\x03 \x01(\x04:\x14\x31\x38\x34\x34\x36\x37\x34\x34\x30\x37\x33\x37\x30\x39\x35\x35\x31\x36\x31\x35\x12\x14\n\tStringLen\x18\x04 \x01(\x04:\x01\x38\x12\x15\n\nColumnsCnt\x18\x05 \x01(\x04:\x01\x32\x12\x12\n\x07RowsCnt\x18\x06 \x01(\x04:\x01\x31\x42\n\n\x08Workload\x1aZ\n\x0bTMemoryLoad\x12\x0b\n\x03Tag\x18\x01 \x01(\x04\x12\x17\n\x0f\x44urationSeconds\x18\x02 \x01(\r\x12\x11\n\tBlockSize\x18\x03 \x01(\x04\x12\x12\n\nIntervalUs\x18\x04 \x01(\x04\">\n\x08\x45LogMode\x12\x10\n\x0cLOG_PARALLEL\x10\x01\x12\x12\n\x0eLOG_SEQUENTIAL\x10\x02\x12\x0c\n\x08LOG_NONE\x10\x03\x42\t\n\x07\x43ommand\"W\n\x13TEvLoadTestResponse\x12\x0e\n\x06Status\x18\x01 \x01(\r\x12\x13\n\x0b\x45rrorReason\x18\x02 \x01(\t\x12\x0e\n\x06\x43ookie\x18\x03 \x01(\x04\x12\x0b\n\x03Tag\x18\x04 \x01(\x04\x42\x18\n\x16ru.yandex.kikimr.proto')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.core.protos.load_test_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026ru.yandex.kikimr.proto'
  _TEVLOADTESTREQUEST._serialized_start=158
  _TEVLOADTESTREQUEST._serialized_end=5494
  _TEVLOADTESTREQUEST_TINTERVALINFO._serialized_start=834
  _TEVLOADTESTREQUEST_TINTERVALINFO._serialized_end=1181
  _TEVLOADTESTREQUEST_TINTERVALINFO_TINTERVALUNIFORM._serialized_start=1025
  _TEVLOADTESTREQUEST_TINTERVALINFO_TINTERVALUNIFORM._serialized_end=1103
  _TEVLOADTESTREQUEST_TINTERVALINFO_TINTERVALPOISSON._serialized_start=1105
  _TEVLOADTESTREQUEST_TINTERVALINFO_TINTERVALPOISSON._serialized_end=1165
  _TEVLOADTESTREQUEST_TSIZEINFO._serialized_start=1183
  _TEVLOADTESTREQUEST_TSIZEINFO._serialized_end=1236
  _TEVLOADTESTREQUEST_TSTORAGELOAD._serialized_start=1239
  _TEVLOADTESTREQUEST_TSTORAGELOAD._serialized_end=2464
  _TEVLOADTESTREQUEST_TSTORAGELOAD_TREQUESTINFO._serialized_start=1459
  _TEVLOADTESTREQUEST_TSTORAGELOAD_TREQUESTINFO._serialized_end=1580
  _TEVLOADTESTREQUEST_TSTORAGELOAD_TTABLETINFO._serialized_start=1583
  _TEVLOADTESTREQUEST_TSTORAGELOAD_TTABLETINFO._serialized_end=1775
  _TEVLOADTESTREQUEST_TSTORAGELOAD_TPERTABLETPROFILE._serialized_start=1778
  _TEVLOADTESTREQUEST_TSTORAGELOAD_TPERTABLETPROFILE._serialized_end=2464
  _TEVLOADTESTREQUEST_TSTOP._serialized_start=2466
  _TEVLOADTESTREQUEST_TSTOP._serialized_end=2509
  _TEVLOADTESTREQUEST_TPDISKWRITELOAD._serialized_start=2512
  _TEVLOADTESTREQUEST_TPDISKWRITELOAD._serialized_end=2973
  _TEVLOADTESTREQUEST_TPDISKWRITELOAD_TCHUNKINFO._serialized_start=2930
  _TEVLOADTESTREQUEST_TPDISKWRITELOAD_TCHUNKINFO._serialized_end=2973
  _TEVLOADTESTREQUEST_TVDISKLOAD._serialized_start=2976
  _TEVLOADTESTREQUEST_TVDISKLOAD._serialized_end=3530
  _TEVLOADTESTREQUEST_TPDISKREADLOAD._serialized_start=3533
  _TEVLOADTESTREQUEST_TPDISKREADLOAD._serialized_end=3915
  _TEVLOADTESTREQUEST_TPDISKREADLOAD_TCHUNKINFO._serialized_start=2930
  _TEVLOADTESTREQUEST_TPDISKREADLOAD_TCHUNKINFO._serialized_end=2973
  _TEVLOADTESTREQUEST_TPDISKLOGLOAD._serialized_start=3918
  _TEVLOADTESTREQUEST_TPDISKLOGLOAD._serialized_end=4347
  _TEVLOADTESTREQUEST_TPDISKLOGLOAD_TWORKERCONFIG._serialized_start=4117
  _TEVLOADTESTREQUEST_TPDISKLOGLOAD_TWORKERCONFIG._serialized_end=4347
  _TEVLOADTESTREQUEST_TKEYVALUELOAD._serialized_start=4350
  _TEVLOADTESTREQUEST_TKEYVALUELOAD._serialized_end=4624
  _TEVLOADTESTREQUEST_TKEYVALUELOAD_TWORKERCONFIG._serialized_start=4503
  _TEVLOADTESTREQUEST_TKEYVALUELOAD_TWORKERCONFIG._serialized_end=4624
  _TEVLOADTESTREQUEST_TKQPLOAD._serialized_start=4627
  _TEVLOADTESTREQUEST_TKQPLOAD._serialized_end=5327
  _TEVLOADTESTREQUEST_TKQPLOAD_TSTOCKWORKLOAD._serialized_start=4988
  _TEVLOADTESTREQUEST_TKQPLOAD_TSTOCKWORKLOAD._serialized_end=5131
  _TEVLOADTESTREQUEST_TKQPLOAD_TKVWORKLOAD._serialized_start=5134
  _TEVLOADTESTREQUEST_TKQPLOAD_TKVWORKLOAD._serialized_end=5315
  _TEVLOADTESTREQUEST_TMEMORYLOAD._serialized_start=5329
  _TEVLOADTESTREQUEST_TMEMORYLOAD._serialized_end=5419
  _TEVLOADTESTREQUEST_ELOGMODE._serialized_start=5421
  _TEVLOADTESTREQUEST_ELOGMODE._serialized_end=5483
  _TEVLOADTESTRESPONSE._serialized_start=5496
  _TEVLOADTESTRESPONSE._serialized_end=5583
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/core/protos/kqp_stats.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb.library.yql.dq.actors.protos import dq_stats_pb2 as ydb_dot_library_dot_yql_dot_dq_dot_actors_dot_protos_dot_dq__stats__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fydb/core/protos/kqp_stats.proto\x12\tNKqpProto\x1a/ydb/library/yql/dq/actors/protos/dq_stats.proto\"L\n\x10TKqpStatsCompile\x12\x11\n\tFromCache\x18\x01 \x01(\x08\x12\x12\n\nDurationUs\x18\x02 \x01(\x04\x12\x11\n\tCpuTimeUs\x18\x03 \x01(\x04\"C\n\x18TKqpShardTableExtraStats\x12\x0f\n\x07ShardId\x18\x01 \x01(\x04\x12\x16\n\x0eShardCpuTimeUs\x18\x02 \x01(\x04\"k\n\x1cTKqpShardTableAggrExtraStats\x12\x16\n\x0e\x41\x66\x66\x65\x63tedShards\x18\x01 \x01(\r\x12\x33\n\x0eShardCpuTimeUs\x18\x02 \x01(\x0b\x32\x1b.NYql.NDqProto.TDqStatsAggr\"\xc7\x01\n\x17TKqpScanTableExtraStats\x12\x18\n\x10IScanStartTimeMs\x18\x01 \x01(\x04\x12\x19\n\x11IScanFinishTimeMs\x18\x02 \x01(\x04\x12\x16\n\x0eIScanCpuTimeUs\x18\x03 \x01(\x04\x12\x17\n\x0fIScanWaitTimeUs\x18\x04 \x01(\x04\x12\x17\n\x0fIScanPageFaults\x18\x05 \x01(\x04\x12\x10\n\x08Messages\x18\x06 \x01(\x04\x12\x1b\n\x13MessagesByPageFault\x18\x07 \x01(\x04\"\x1c\n\x1aTKqpComputeActorExtraStats\"J\n\x13TKqpStageExtraStats\x12\x33\n\x0e\x44\x61tashardTasks\x18\x01 \x03(\x0b\x32\x1b.NYql.NDqProto.TDqTaskStats\"\xd3\x01\n\x17TKqpExecutionExtraStats\x12\x16\n\x0e\x41\x66\x66\x65\x63tedShards\x18\x01 \x01(\r\x12\x35\n\x10\x43omputeCpuTimeUs\x18\x02 \x01(\x0b\x32\x1b.NYql.NDqProto.TDqStatsAggr\x12\x34\n\x0fShardsCpuTimeUs\x18\x03 \x01(\x0b\x32\x1b.NYql.NDqProto.TDqStatsAggr\x12\x18\n\x10ResolveCpuTimeUs\x18\x64 \x01(\x04\x12\x19\n\x11ResolveWallTimeUs\x18\x65 \x01(\x04\"\xfa\x01\n\x0eTKqpStatsQuery\x12\x12\n\nDurationUs\x18\x01 \x01(\x04\x12\x30\n\x0b\x43ompilation\x18\x02 \x01(\x0b\x32\x1b.NKqpProto.TKqpStatsCompile\x12\x17\n\x0fWorkerCpuTimeUs\x18\x04 \x01(\x04\x12\x15\n\rReadSetsCount\x18\x05 \x01(\x04\x12\x1b\n\x13MaxShardProgramSize\x18\x06 \x01(\x04\x12\x19\n\x11MaxShardReplySize\x18\x07 \x01(\x04\x12\x34\n\nExecutions\x18\x08 \x03(\x0b\x32 .NYql.NDqProto.TDqExecutionStatsJ\x04\x08\x03\x10\x04\x42\x1b\n\x16ru.yandex.kikimr.proto\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.core.protos.kqp_stats_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026ru.yandex.kikimr.proto\370\001\001'
  _TKQPSTATSCOMPILE._serialized_start=95
  _TKQPSTATSCOMPILE._serialized_end=171
  _TKQPSHARDTABLEEXTRASTATS._serialized_start=173
  _TKQPSHARDTABLEEXTRASTATS._serialized_end=240
  _TKQPSHARDTABLEAGGREXTRASTATS._serialized_start=242
  _TKQPSHARDTABLEAGGREXTRASTATS._serialized_end=349
  _TKQPSCANTABLEEXTRASTATS._serialized_start=352
  _TKQPSCANTABLEEXTRASTATS._serialized_end=551
  _TKQPCOMPUTEACTOREXTRASTATS._serialized_start=553
  _TKQPCOMPUTEACTOREXTRASTATS._serialized_end=581
  _TKQPSTAGEEXTRASTATS._serialized_start=583
  _TKQPSTAGEEXTRASTATS._serialized_end=657
  _TKQPEXECUTIONEXTRASTATS._serialized_start=660
  _TKQPEXECUTIONEXTRASTATS._serialized_end=871
  _TKQPSTATSQUERY._serialized_start=874
  _TKQPSTATSQUERY._serialized_end=1124
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/core/protos/statestorage.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from library.cpp.actors.protos import actors_pb2 as library_dot_cpp_dot_actors_dot_protos_dot_actors__pb2
from ydb.core.protos import base_pb2 as ydb_dot_core_dot_protos_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"ydb/core/protos/statestorage.proto\x12\x13NKikimrStateStorage\x1a&library/cpp/actors/protos/actors.proto\x1a\x1aydb/core/protos/base.proto\"-\n\tTEvLookup\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\x12\x0e\n\x06\x43ookie\x18\x02 \x01(\x04\"\xca\x03\n\x07TEvInfo\x12*\n\x06Status\x18\x01 \x01(\x0e\x32\x1a.NKikimrProto.EReplyStatus\x12\x10\n\x08TabletID\x18\x02 \x01(\x06\x12\x0e\n\x06\x43ookie\x18\x03 \x01(\x04\x12-\n\rCurrentLeader\x18\x04 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x33\n\x13\x43urrentLeaderTablet\x18\x05 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x19\n\x11\x43urrentGeneration\x18\x06 \x01(\r\x12\x13\n\x0b\x43urrentStep\x18\x07 \x01(\r\x12\x0e\n\x06Locked\x18\x08 \x01(\x08\x12\x11\n\tLockedFor\x18\t \x01(\x04\x12\x11\n\tSignature\x18\n \x01(\x04\x12(\n\x08\x46ollower\x18\x0b \x03(\x0b\x32\x16.NActorsProto.TActorId\x12.\n\x0e\x46ollowerTablet\x18\x0c \x03(\x0b\x32\x16.NActorsProto.TActorId\x12\x32\n\x12\x46ollowerCandidates\x18\r \x03(\x0b\x32\x16.NActorsProto.TActorId\x12\x19\n\x11\x43onfigContentHash\x18\x0e \x01(\x07\"\x14\n\x12TEvReplicaShutdown\"\x10\n\x0eTEvDumpRequest\"5\n\x07TEvDump\x12*\n\x04Info\x18\x01 \x03(\x0b\x32\x1c.NKikimrStateStorage.TEvInfo\"\xec\x01\n\tTEvUpdate\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\x12\x0e\n\x06\x43ookie\x18\x02 \x01(\x04\x12.\n\x0eProposedLeader\x18\x03 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x34\n\x14ProposedLeaderTablet\x18\x04 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x1a\n\x12ProposedGeneration\x18\x05 \x01(\r\x12\x14\n\x0cProposedStep\x18\x06 \x01(\r\x12\x11\n\tSignature\x18\x07 \x01(\x04\x12\x12\n\nIsGuardian\x18\x08 \x01(\x08\"\x1d\n\tTEvDelete\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\"N\n\nTEvCleanup\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\x12.\n\x0eProposedLeader\x18\x02 \x01(\x0b\x32\x16.NActorsProto.TActorId\"\x94\x01\n\x13TEvRegisterFollower\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\x12(\n\x08\x46ollower\x18\x02 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12.\n\x0e\x46ollowerTablet\x18\x03 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x11\n\tCandidate\x18\x04 \x01(\x08\"S\n\x15TEvUnregisterFollower\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\x12(\n\x08\x46ollower\x18\x02 \x01(\x0b\x32\x16.NActorsProto.TActorId\"\x8a\x01\n\x07TEvLock\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\x12\x0e\n\x06\x43ookie\x18\x02 \x01(\x04\x12.\n\x0eProposedLeader\x18\x03 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x1a\n\x12ProposedGeneration\x18\x04 \x01(\r\x12\x11\n\tSignature\x18\x05 \x01(\x04\">\n\x17TEvReplicaLeaderDemoted\x12\x10\n\x08TabletID\x18\x01 \x01(\x06\x12\x11\n\tSignature\x18\x02 \x01(\x04\"\x7f\n\x16TEvReplicaBoardPublish\x12\x0c\n\x04Path\x18\x01 \x01(\t\x12\x0f\n\x07Payload\x18\x02 \x01(\x0c\x12\r\n\x05TtlMs\x18\x03 \x01(\x04\x12\x10\n\x08Register\x18\x04 \x01(\x08\x12%\n\x05Owner\x18\x05 \x01(\x0b\x32\x16.NActorsProto.TActorId\"\x1b\n\x19TEvReplicaBoardPublishAck\"_\n\x15TEvReplicaBoardLookup\x12\x0c\n\x04Path\x18\x01 \x01(\t\x12%\n\x05Owner\x18\x02 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x11\n\tSubscribe\x18\x03 \x01(\x08\"\x18\n\x16TEvReplicaBoardCleanup\"I\n\x0fTBoardEntryInfo\x12%\n\x05Owner\x18\x01 \x01(\x0b\x32\x16.NActorsProto.TActorId\x12\x0f\n\x07Payload\x18\x02 \x01(\x0c\"h\n\x13TEvReplicaBoardInfo\x12\x0c\n\x04Path\x18\x01 \x01(\t\x12\x0f\n\x07\x44ropped\x18\x02 \x01(\x08\x12\x32\n\x04Info\x18\x03 \x03(\x0b\x32$.NKikimrStateStorage.TBoardEntryInfo\"\xcb\x01\n\x13TEndpointBoardEntry\x12\x0f\n\x07\x41\x64\x64ress\x18\x01 \x01(\t\x12\x0c\n\x04Port\x18\x02 \x01(\r\x12\x0c\n\x04Load\x18\x03 \x01(\x02\x12\x0b\n\x03Ssl\x18\x04 \x01(\x08\x12\x10\n\x08Services\x18\x05 \x03(\t\x12\x12\n\nDataCenter\x18\x06 \x01(\t\x12\x0e\n\x06NodeId\x18\x07 \x01(\r\x12\x13\n\x0b\x41\x64\x64ressesV4\x18\x08 \x03(\t\x12\x13\n\x0b\x41\x64\x64ressesV6\x18\t \x03(\t\x12\x1a\n\x12TargetNameOverride\x18\n \x01(\tB\x18\n\x16ru.yandex.kikimr.proto')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.core.protos.statestorage_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026ru.yandex.kikimr.proto'
  _TEVLOOKUP._serialized_start=127
  _TEVLOOKUP._serialized_end=172
  _TEVINFO._serialized_start=175
  _TEVINFO._serialized_end=633
  _TEVREPLICASHUTDOWN._serialized_start=635
  _TEVREPLICASHUTDOWN._serialized_end=655
  _TEVDUMPREQUEST._serialized_start=657
  _TEVDUMPREQUEST._serialized_end=673
  _TEVDUMP._serialized_start=675
  _TEVDUMP._serialized_end=728
  _TEVUPDATE._serialized_start=731
  _TEVUPDATE._serialized_end=967
  _TEVDELETE._serialized_start=969
  _TEVDELETE._serialized_end=998
  _TEVCLEANUP._serialized_start=1000
  _TEVCLEANUP._serialized_end=1078
  _TEVREGISTERFOLLOWER._serialized_start=1081
  _TEVREGISTERFOLLOWER._serialized_end=1229
  _TEVUNREGISTERFOLLOWER._serialized_start=1231
  _TEVUNREGISTERFOLLOWER._serialized_end=1314
  _TEVLOCK._serialized_start=1317
  _TEVLOCK._serialized_end=1455
  _TEVREPLICALEADERDEMOTED._serialized_start=1457
  _TEVREPLICALEADERDEMOTED._serialized_end=1519
  _TEVREPLICABOARDPUBLISH._serialized_start=1521
  _TEVREPLICABOARDPUBLISH._serialized_end=1648
  _TEVREPLICABOARDPUBLISHACK._serialized_start=1650
  _TEVREPLICABOARDPUBLISHACK._serialized_end=1677
  _TEVREPLICABOARDLOOKUP._serialized_start=1679
  _TEVREPLICABOARDLOOKUP._serialized_end=1774
  _TEVREPLICABOARDCLEANUP._serialized_start=1776
  _TEVREPLICABOARDCLEANUP._serialized_end=1800
  _TBOARDENTRYINFO._serialized_start=1802
  _TBOARDENTRYINFO._serialized_end=1875
  _TEVREPLICABOARDINFO._serialized_start=1877
  _TEVREPLICABOARDINFO._serialized_end=1981
  _TENDPOINTBOARDENTRY._serialized_start=1984
  _TENDPOINTBOARDENTRY._serialized_end=2187
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/public/api/protos/ydb_clickhouse_internal.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb.public.api.protos import ydb_operation_pb2 as ydb_dot_public_dot_api_dot_protos_dot_ydb__operation__pb2
from ydb.public.api.protos import ydb_table_pb2 as ydb_dot_public_dot_api_dot_protos_dot_ydb__table__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3ydb/public/api/protos/ydb_clickhouse_internal.proto\x12\x16Ydb.ClickhouseInternal\x1a)ydb/public/api/protos/ydb_operation.proto\x1a%ydb/public/api/protos/ydb_table.proto\"\xfa\x01\n\x0bScanRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\r\n\x05table\x18\x02 \x01(\t\x12\x0f\n\x07\x63olumns\x18\x03 \x03(\t\x12\x10\n\x08\x66rom_key\x18\x04 \x01(\x0c\x12\x1a\n\x12\x66rom_key_inclusive\x18\x05 \x01(\x08\x12\x0e\n\x06to_key\x18\x06 \x01(\x0c\x12\x18\n\x10to_key_inclusive\x18\x07 \x01(\x08\x12\x10\n\x08max_rows\x18\x08 \x01(\x04\x12\x11\n\tmax_bytes\x18\t \x01(\x04\x12\x13\n\x0bsnapshot_id\x18\n \x01(\t\"<\n\x0cScanResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"W\n\nScanResult\x12\x0e\n\x06\x62locks\x18\x01 \x03(\x0c\x12\x0b\n\x03\x65os\x18\x02 \x01(\x08\x12\x10\n\x08last_key\x18\x03 \x01(\x0c\x12\x1a\n\x12last_key_inclusive\x18\x04 \x01(\x08\"i\n\x18GetShardLocationsRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x12\n\ntablet_ids\x18\x02 \x03(\x04\"I\n\x19GetShardLocationsResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\";\n\nTabletInfo\x12\x11\n\ttablet_id\x18\x01 \x01(\x04\x12\x0c\n\x04host\x18\x02 \x01(\t\x12\x0c\n\x04port\x18\x03 \x01(\r\"N\n\x17GetShardLocationsResult\x12\x33\n\x07tablets\x18\x01 \x03(\x0b\x32\".Ydb.ClickhouseInternal.TabletInfo\"\x80\x01\n\x14\x44\x65scribeTableRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x1f\n\x17include_partitions_info\x18\x03 \x01(\x08\"N\n\rPartitionInfo\x12\x11\n\ttablet_id\x18\x01 \x01(\x04\x12\x0f\n\x07\x65nd_key\x18\x02 \x01(\x0c\x12\x19\n\x11\x65nd_key_inclusive\x18\x03 \x01(\x08\"\xa1\x01\n\x13\x44\x65scribeTableResult\x12&\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x15.Ydb.Table.ColumnMeta\x12\x13\n\x0bprimary_key\x18\x03 \x03(\t\x12\x39\n\npartitions\x18\x04 \x03(\x0b\x32%.Ydb.ClickhouseInternal.PartitionInfo\x12\x12\n\nis_virtual\x18\x05 \x01(\x08\"E\n\x15\x44\x65scribeTableResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"}\n\x15\x43reateSnapshotRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x03(\t\x12\x1b\n\x13ignore_system_views\x18\x03 \x01(\x08\"?\n\x14\x43reateSnapshotResult\x12\x13\n\x0bsnapshot_id\x18\x01 \x01(\t\x12\x12\n\ntimeout_ms\x18\x02 \x01(\x04\"F\n\x16\x43reateSnapshotResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x93\x01\n\x16RefreshSnapshotRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x03(\t\x12\x13\n\x0bsnapshot_id\x18\x03 \x01(\t\x12\x1b\n\x13ignore_system_views\x18\x04 \x01(\x08\",\n\x15RefreshSnapshotResult\x12\x13\n\x0bsnapshot_id\x18\x01 \x01(\t\"G\n\x17RefreshSnapshotResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x93\x01\n\x16\x44iscardSnapshotRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x03(\t\x12\x13\n\x0bsnapshot_id\x18\x03 \x01(\t\x12\x1b\n\x13ignore_system_views\x18\x04 \x01(\x08\",\n\x15\x44iscardSnapshotResult\x12\x13\n\x0bsnapshot_id\x18\x01 \x01(\t\"G\n\x17\x44iscardSnapshotResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.OperationB8\n\x19\x63om.yandex.ydb.clickhouseB\x18\x43lickhouseInternalProtos\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.public.api.protos.ydb_clickhouse_internal_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.yandex.ydb.clickhouseB\030ClickhouseInternalProtos\370\001\001'
  _SCANREQUEST._serialized_start=162
  _SCANREQUEST._serialized_end=412
  _SCANRESPONSE._serialized_start=414
  _SCANRESPONSE._serialized_end=474
  _SCANRESULT._serialized_start=476
  _SCANRESULT._serialized_end=563
  _GETSHARDLOCATIONSREQUEST._serialized_start=565
  _GETSHARDLOCATIONSREQUEST._serialized_end=670
  _GETSHARDLOCATIONSRESPONSE._serialized_start=672
  _GETSHARDLOCATIONSRESPONSE._serialized_end=745
  _TABLETINFO._serialized_start=747
  _TABLETINFO._serialized_end=806
  _GETSHARDLOCATIONSRESULT._serialized_start=808
  _GETSHARDLOCATIONSRESULT._serialized_end=886
  _DESCRIBETABLEREQUEST._serialized_start=889
  _DESCRIBETABLEREQUEST._serialized_end=1017
  _PARTITIONINFO._serialized_start=1019
  _PARTITIONINFO._serialized_end=1097
  _DESCRIBETABLERESULT._serialized_start=1100
  _DESCRIBETABLERESULT._serialized_end=1261
  _DESCRIBETABLERESPONSE._serialized_start=1263
  _DESCRIBETABLERESPONSE._serialized_end=1332
  _CREATESNAPSHOTREQUEST._serialized_start=1334
  _CREATESNAPSHOTREQUEST._serialized_end=1459
  _CREATESNAPSHOTRESULT._serialized_start=1461
  _CREATESNAPSHOTRESULT._serialized_end=1524
  _CREATESNAPSHOTRESPONSE._serialized_start=1526
  _CREATESNAPSHOTRESPONSE._serialized_end=1596
  _REFRESHSNAPSHOTREQUEST._serialized_start=1599
  _REFRESHSNAPSHOTREQUEST._serialized_end=1746
  _REFRESHSNAPSHOTRESULT._serialized_start=1748
  _REFRESHSNAPSHOTRESULT._serialized_end=1792
  _REFRESHSNAPSHOTRESPONSE._serialized_start=1794
  _REFRESHSNAPSHOTRESPONSE._serialized_end=1865
  _DISCARDSNAPSHOTREQUEST._serialized_start=1868
  _DISCARDSNAPSHOTREQUEST._serialized_end=2015
  _DISCARDSNAPSHOTRESULT._serialized_start=2017
  _DISCARDSNAPSHOTRESULT._serialized_end=2061
  _DISCARDSNAPSHOTRESPONSE._serialized_start=2063
  _DISCARDSNAPSHOTRESPONSE._serialized_end=2134
# @@protoc_insertion_point(module_scope)



import ydb.apps.dstool.lib.common as common
import time
import random
import subprocess
import ydb.apps.dstool.lib.grouptool as grouptool
from datetime import datetime, timedelta
from collections import defaultdict
import sys

description = 'Create workload to stress failure model'


def add_options(p):
    p.add_argument('--disable-wipes', action='store_true', help='Disable VDisk wipes')
    p.add_argument('--disable-evicts', action='store_true', help='Disable VDisk evicts')
    p.add_argument('--disable-restarts', action='store_true', help='Disable node restarts')
    p.add_argument('--enable-pdisk-encryption-keys-changes', action='store_true', help='Enable changes of PDisk encryption keys')


def fetch_start_time_map(base_config):
    start_time_map = {}
    for node_id in {pdisk.NodeId for pdisk in base_config.PDisk}:
        r = common.fetch_json_info('sysinfo', [node_id])
        if len(r) != 1:
            return None
        k, v = r.popitem()
        assert k == node_id
        if 'StartTime' not in v:
            return None
        start_time_map[node_id] = int(v['StartTime'])
    return start_time_map


def make_pdisk_key_config(pdisk_keys, node_id):
    s = ""
    for key in pdisk_keys[node_id]:
        s += "Keys {" + "\n"
        s += "  ContainerPath: " + "\\\"" + key["path"] + "\\\"" + "\n"
        s += "  Pin: " + "\\\"" + key["pin"] + "\\\"" + "\n"
        s += "  Id: " + "\\\"" + key["id"] + "\\\"" + "\n"
        s += "  Version: " + str(key["version"]) + "\n"
        s += "}" + "\n"
    return s


def remove_old_pdisk_keys(pdisk_keys, pdisk_key_versions, node_id):
    v = pdisk_key_versions[node_id]
    for pdisk_key in pdisk_keys[node_id]:
        if pdisk_key["version"] != v:
            pdisk_keys[node_id].remove(pdisk_key)


def update_pdisk_key_config(node_fqdn_map, pdisk_keys, node_id):
    host = node_fqdn_map[node_id]
    subprocess.run('''ssh {0} "sudo echo '{1}' > /Berkanavt/kikimr/cfg/pdisk_key.txt"'''.format(host, make_pdisk_key_config(pdisk_keys, node_id)), shell=True)
    for key in pdisk_keys[node_id]:
        if (len(key["path"]) > 0):
            subprocess.run('''ssh {0} "echo '{1}' | sudo tee {2} >/dev/null"'''.format(host, key["file"], key["path"]), shell=True)


def do(args):
    recent_restarts = []

    pdisk_keys = {}
    pdisk_key_versions = {}

    config_retries = None

    while True:
        common.flush_cache()

        try:
            base_config = common.fetch_base_config()
            vslot_map = common.build_vslot_map(base_config)
            node_fqdn_map = common.build_node_fqdn_map(base_config)
            start_time_map = fetch_start_time_map(base_config)
        except Exception:
            if config_retries is None:
                config_retries = 3
            elif config_retries == 0:
                raise
            else:
                config_retries -= 1
            continue

        config_retries = None

        for vslot in base_config.VSlot:
            assert not vslot.Ready or vslot.Status == 'READY'

        if (len(pdisk_keys) == 0):
            # initialize pdisk_keys and pdisk_key_versions
            for node_id in {pdisk.NodeId for pdisk in base_config.PDisk}:
                pdisk_key_versions[node_id] = 1
                pdisk_keys[node_id] = [{"path" : "", "pin" : "", "id" : "0", "version" : 0, "file" : ""}]

        vdisk_status = defaultdict(lambda: False)
        error = False
        for vslot_id, vdisk in common.fetch_json_info('vdiskinfo').items():
            try:
                key = *vslot_id, *common.get_vdisk_id_json(vdisk['VDiskId'])
                vdisk_status[key] = vdisk['Replicated'] and vdisk['VDiskState'] == 'OK'
            except KeyError:
                common.print_if_not_quiet(args, 'Failed to fetch VDisk status for VSlotId %s' % vslot_id, file=sys.stderr)
                error = True
        if error:
            common.print_if_not_quiet(args, 'Waiting for the next round...', file=sys.stdout)
            time.sleep(1)
            continue

        def can_act_on_vslot(node_id, pdisk_id=None, vslot_id=None):
            def match(x):
                return node_id == x[0] and pdisk_id in [None, x[1]] and vslot_id in [None, x[2]]

            for group in base_config.Group:
                if any(map(match, map(common.get_vslot_id, group.VSlotId))):
                    content = {
                        common.get_vdisk_id_short(vslot): not match(vslot_id) and vslot.Ready and vdisk_status[vslot_id + common.get_vdisk_id(vslot)]
                        for vslot_id in map(common.get_vslot_id, group.VSlotId)
                        for vslot in [vslot_map[vslot_id]]
                    }
                    common.print_if_verbose(args, content, file=sys.stderr)
                    if not grouptool.check_fail_model(content, group.ErasureSpecies):
                        return False
            return True

        def do_restart(node_id):
            host = node_fqdn_map[node_id]
            if args.enable_pdisk_encryption_keys_changes:
                update_pdisk_key_config(node_fqdn_map, pdisk_keys, node_id)
            subprocess.call(['ssh', host, 'sudo', 'killall', '-9', 'kikimr'])
            if args.enable_pdisk_encryption_keys_changes:
                remove_old_pdisk_keys(pdisk_keys, pdisk_key_versions, node_id)

        def do_evict(vslot_id):
            assert can_act_on_vslot(*vslot_id)
            try:
                request = common.kikimr_bsconfig.TConfigRequest(IgnoreDegradedGroupsChecks=True)
                vslot = vslot_map[vslot_id]
                cmd = request.Command.add().ReassignGroupDisk
                cmd.GroupId = vslot.GroupId
                cmd.GroupGeneration = vslot.GroupGeneration
                cmd.FailRealmIdx = vslot.FailRealmIdx
                cmd.FailDomainIdx = vslot.FailDomainIdx
                cmd.VDiskIdx = vslot.VDiskIdx
                response = common.invoke_bsc_request(request)
                if not response.Success:
                    if 'Error# failed to allocate group: no group options' in response.ErrorDescription:
                        common.print_if_verbose(args, response)
                    else:
                        raise Exception('Unexpected error from BSC: %s' % response.ErrorDescription)
            except Exception as e:
                raise Exception('Failed to perform evict request: %s' % e)

        def do_wipe(vslot):
            assert can_act_on_vslot(*common.get_vslot_id(vslot.VSlotId))
            try:
                request = common.create_wipe_request(args, vslot)
                common.invoke_wipe_request(request)
            except Exception as e:
                raise Exception('Failed to perform wipe request: %s' % e)

        def do_add_pdisk_key(node_id):
            pdisk_key_versions[node_id] += 1
            v = pdisk_key_versions[node_id]
            pdisk_keys[node_id].append({"path" : "/Berkanavt/kikimr/cfg/pdisk_key_" + str(v) + ".txt",
                                        "pin" : "",
                                        "id" : "Key" + str(v),
                                        "version" : v,
                                        "file" : "keynumber" + str(v)})

        ################################################################################################################

        now = datetime.utcnow()
        while recent_restarts and recent_restarts[0] + timedelta(minutes=1) < now:
            recent_restarts.pop(0)

        possible_actions = []

        for vslot in base_config.VSlot:
            if common.is_dynamic_group(vslot.GroupId):
                vslot_id = common.get_vslot_id(vslot.VSlotId)
                if can_act_on_vslot(*vslot_id) and recent_restarts:
                    vdisk_id = '[%08x:%d:%d:%d]' % (vslot.GroupId, vslot.FailRealmIdx, vslot.FailDomainIdx, vslot.VDiskIdx)
                    if not args.disable_evicts:
                        possible_actions.append(('evict vslot id: %s, vdisk id: %s' % (vslot_id, vdisk_id), (do_evict, vslot_id)))
                    if not args.disable_wipes:
                        possible_actions.append(('wipe vslot id: %s, vdisk id: %s' % (vslot_id, vdisk_id), (do_wipe, vslot)))

        if start_time_map and len(recent_restarts) < 3:
            # sort so that the latest restarts come first
            nodes_to_restart = sorted(start_time_map, key=start_time_map.__getitem__)
            node_count = len(nodes_to_restart)
            nodes_to_restart = nodes_to_restart[:node_count//2]
            for node_id in nodes_to_restart:
                if args.enable_pdisk_encryption_keys_changes:
                    possible_actions.append(('add new pdisk key to node with id: %d' % node_id, (do_add_pdisk_key, node_id)))
                if not args.disable_restarts:
                    possible_actions.append(('restart node with id: %d' % node_id, (do_restart, node_id)))

        if not possible_actions:
            common.print_if_not_quiet(args, 'Waiting for the next round...', file=sys.stdout)
            time.sleep(1)
            continue

        ################################################################################################################

        action_name, action = random.choice(possible_actions)
        common.print_if_not_quiet(args, '%s' % action_name, file=sys.stdout)

        try:
            action[0](*action[1:])
            if action_name.startswith('restart'):
                recent_restarts.append(now)
        except Exception as e:
            common.print_if_not_quiet(args, 'Failed to perform action: %s with error: %s' % (action_name, e), file=sys.stderr)

        common.print_if_not_quiet(args, 'Waiting for the next round...', file=sys.stdout)
        time.sleep(1)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb/core/protos/blobstorage_disk_color.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,ydb/core/protos/blobstorage_disk_color.proto\x12\x12NKikimrBlobStorage\"\x8c\x01\n\x10TPDiskSpaceColor\"x\n\x01\x45\x12\t\n\x05GREEN\x10\x00\x12\x08\n\x04\x43YAN\x10\n\x12\x10\n\x0cLIGHT_YELLOW\x10\x0f\x12\n\n\x06YELLOW\x10\x14\x12\x10\n\x0cLIGHT_ORANGE\x10\x1e\x12\x0e\n\nPRE_ORANGE\x10#\x12\n\n\x06ORANGE\x10(\x12\x07\n\x03RED\x10\x32\x12\t\n\x05\x42LACK\x10<b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb.core.protos.blobstorage_disk_color_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TPDISKSPACECOLOR._serialized_start=69
  _TPDISKSPACECOLOR._serialized_end=209
  _TPDISKSPACECOLOR_E._serialized_start=89
  _TPDISKSPACECOLOR_E._serialized_end=209
# @@protoc_insertion_point(module_scope)



from typing import NamedTuple
from collections import defaultdict
from functools import reduce
from operator import getitem, itemgetter
from itertools import product
import ydb.apps.dstool.lib.common as common
import re
import sys


def decompose_location_map_by_levels(levels, locations):
    if levels == (0,) * len(levels):
        levels = 10, 20, 10, 40
    id_map = {}

    def get_id(fdom, begin, end):
        return id_map.setdefault(fdom.subs(begin, end), len(id_map) + 1)
    return {
        location: PDiskLocation(
            get_id(location, 0, levels[0]),
            get_id(location, levels[0], levels[1]),
            get_id(location, 0, levels[2]),
            get_id(location, levels[2], levels[3]),
        )
        for location in locations
    }


def decompose_location_map(sp, locations):
    g = sp.Geometry
    levels = g.RealmLevelBegin, g.RealmLevelEnd, g.DomainLevelBegin, g.DomainLevelEnd
    return decompose_location_map_by_levels(levels, locations)


def check_group(content, location_id_map=None):
    main_map = defaultdict(set)
    prefix_map = defaultdict(lambda: defaultdict(set))
    for coord, location in content:
        if location_id_map is not None:
            location = location_id_map[location]
        for coord_depth, location_depth in [(0, 1), (1, 2), (1, 3), (2, 4)]:
            key = coord[:coord_depth]
            value = location[:location_depth]
            main_map[key].add(value)
            if key:
                prefix_map[key[:-1]][key[-1]].add(value)
    errs = []
    for key, value in main_map.items():
        if len(value) != len(set(map(len, value))):
            errs.append('nonunique locations at key "%s": %d != %d' % (':'.join(map(str, key)), len(value), len(set(map(len, value)))))
    for key, subvalues in ((k, list(map(frozenset, v.values()))) for k, v in prefix_map.items()):
        if len(set(subvalues)) != len(subvalues):
            errs.append('intersecting locations at prefix "%s": %d != %d' % (':'.join(map(str, key)), len(set(subvalues)), len(subvalues)))
    return ', '.join(errs) if errs else None


def check_fail_model(coord_to_status, erasure):
    status_per_fdom = {}
    for key, value in coord_to_status.items():
        status_per_fdom[key[:2]] = status_per_fdom.get(key[:2], True) and value
    if erasure == 'none':
        return all(status_per_fdom.values())
    elif erasure in ['block-4-2', 'mirror-3of4']:
        return sum(status_per_fdom.values()) >= len(status_per_fdom) - 2
    elif erasure == 'mirror-3-dc':
        nwdom = defaultdict(int)
        for (fr, fd), status in status_per_fdom.items():
            if not status:
                nwdom[fr] += 1
        return len(nwdom) <= 1 or (len(nwdom) == 2 and any(v == 1 for v in nwdom.values()))
    elif erasure == 'mirror-3':
        return sum(status_per_fdom.values()) >= len(status_per_fdom) - 1
    else:
        assert False, 'unexpected erasure type %s' % erasure


def vdisk_id_from_json(j):
    return itemgetter('GroupID', 'GroupGeneration', 'Ring', 'Domain', 'VDisk')(j)


def vdisk_id_to_string(vdisk_id):
    return '[%08x:%u:%u:%u:%u]' % vdisk_id


class PDiskLocation(NamedTuple):
    realm_prefix: int
    realm_infix: int
    domain_prefix: int
    domain_infix: int

    def __str__(self):
        return ','.join(map(str, self))

    def __repr__(self):
        return 'PDiskLocation(%s)' % ', '.join(map(str, self))


class PDisk(object):
    def __init__(self, pdisk, location, num_used_slots, expected_slot_count, groups):
        assert num_used_slots >= 0
        self.pdisk = pdisk
        self.location = location
        self.num_used_slots = num_used_slots
        self.expected_slot_count = expected_slot_count
        self.groups = groups

    def __str__(self):
        return '%d:%d@%s(%d/%d)/%s' % (self.pdisk.NodeId, self.pdisk.PDiskId, self.location, self.num_used_slots,
                                       self.expected_slot_count, self.groups)

    def __repr__(self):
        return str(self)

    def get_id(self):
        return self.pdisk.NodeId, self.pdisk.PDiskId

    def fits(self, max_num_slots):
        return self.num_used_slots + 1 <= self.expected_slot_count and self.num_used_slots <= max_num_slots


class GroupMapper(object):
    _geom_for_erasure = {
        'mirror-3-dc': (3, 3, 1),
        'block-4-2': (1, 8, 1),
        'mirror-3': (1, 4, 1),
        'none': (1, 1, 1),
        'mirror-3of4': (1, 8, 1),
    }

    def __init__(self, pdisk_map, sp):
        self.pdisk_map = pdisk_map
        self.sp = sp
        self.num_fr, self.num_fdom_in_fr, self.num_vdisks_in_fdom = self.get_geometry()
        self.group_id = 1 << 32

    def create_group(self, content=None):
        group = [[[None for _ in range(self.num_vdisks_in_fdom)] for _ in range(self.num_fdom_in_fr)] for _ in range(self.num_fr)]
        if content is not None:
            for coord, value in content:
                reduce(getitem, coord[:-1], group)[coord[-1]] = value
        return group

    def allocate(self, existing_groups, randomize, honor_existing_groups, group_slot_size, pdisk_free_space):
        for max_num_slots in range(0, 256):
            def reduce_dict(d, threshold):
                res = defaultdict(int)
                for prefix, num_items in d.items():
                    if num_items >= threshold:
                        res[prefix[:-1]] += 1
                return res

            # calculate maximum number of slots suitable for this group
            d = defaultdict(int)
            for disk in self.pdisk_map.values():
                if disk.fits(max_num_slots):
                    d[disk.location[:]] += 1

            # reduce maps
            d = reduce_dict(d, self.num_vdisks_in_fdom)
            d = reduce_dict(d, self.num_fdom_in_fr)
            d = reduce_dict(d, 1)
            d = reduce_dict(d, self.num_fr)
            if d:
                break
        else:
            raise Exception('unable to allocate new group, sorry')

        usable_disks = [
            disk
            for disk in self.pdisk_map.values()
            if disk.fits(max_num_slots)
        ]
        num_matching_disks_in_domain = defaultdict(int)
        for disk in usable_disks:
            num_matching_disks_in_domain[disk.location] += 1

        def remove_disks_with_prefix(prefix):
            prefix_len = len(prefix)
            usable_disks[:] = [
                disk
                for disk in usable_disks
                if disk.location[:prefix_len] != prefix
            ]
            return prefix[:-1]  # drop the last item of the prefix

        group = self.create_group()

        initial_disk_score = {
            pdisk1.get_id(): max(len(pdisk1.groups & pdisk2.groups & existing_groups) for pdisk2 in usable_disks)
            for pdisk1 in usable_disks
        }

        num_disks_in_common = defaultdict(int)  # indexed with neighbor group id
        first = True

        def get_score(disk):
            score = disk.num_used_slots,
            if first:
                score += -initial_disk_score[disk.get_id()],
            common_existing = max((num_disks_in_common[group_id] for group_id in disk.groups if group_id in existing_groups), default=0)
            if honor_existing_groups:
                score += -common_existing,
            common_other = max((num_disks_in_common[group_id] for group_id in disk.groups if group_id not in existing_groups), default=0)
            if randomize:
                score += common_other,
            else:
                score += -common_other,
            if not honor_existing_groups:
                score += -common_existing,
            score += disk.get_id(),
            return score

        subset_disks = {}

        # enumerate allowed disks for specific realm and domain
        def get_allowed_disks(ri, di):
            for key in [(ri, di), (ri,), ()]:
                if key in subset_disks:
                    res = subset_disks[key]
                    break
            else:
                res = usable_disks
            assert res
            return res

        while True:
            options = (
                (disk, ri, di, vi)
                for ri, di, vi in product(range(self.num_fr), range(self.num_fdom_in_fr), range(self.num_vdisks_in_fdom))
                if group[ri][di][vi] is None
                for disk in get_allowed_disks(ri, di)
            )
            res = min(options, key=lambda x: get_score(x[0]), default=None)
            if res is None:
                break
            disk, ri, di, vi = res

            num_matching_disks_in_domain[disk.location] -= 1
            disk.num_used_slots += 1  # mark this slot as used one
            disk.groups.add(self.group_id)  # add just created group to group map

            # adjust number of disks-in-common
            for group_id in disk.groups:
                num_disks_in_common[group_id] += 1

            # store disk in the group
            group[ri][di][vi] = disk.get_id()

            # update disk sets
            base = usable_disks
            for key, prefix_len in [((), 1), ((ri,), 2), ((ri, di), 4)]:
                if key not in subset_disks:
                    prefix = disk.location[:prefix_len]
                    subset_disks[key] = list(filter(lambda disk: disk.location[:prefix_len] == prefix, base))
                    base[:] = list(filter(lambda disk: disk.location[:prefix_len] != prefix, base))
                base = subset_disks[key]

            # remove selected disk from realm/domain disk set
            base[:] = list(filter(lambda x: x.get_id() != disk.get_id(), base))

            first = False

        for existing_group_id in sorted(existing_groups, key=lambda group_id: num_disks_in_common[group_id], reverse=True):
            all_disks = sum(sum(group, []), [])
            if all(group_slot_size[existing_group_id] <= pdisk_free_space[pdisk_id] for pdisk_id in all_disks):
                group_slot_size[self.group_id] = group_slot_size[existing_group_id]
                self.group_id += 1
                return group, existing_group_id

        assert False, 'failed to allocate group'

    def get_geometry(self):
        g = self.sp.Geometry
        geom = g.NumFailRealms, g.NumFailDomainsPerFailRealm, g.NumVDisksPerFailDomain
        if not any(geom):
            geom = self._geom_for_erasure[self.sp.ErasureSpecies]
        return geom


table_re = re.compile('<tbody>(.*?)</tbody>')
row_re = re.compile('<tr>(.*?)</tr>')
cell_re = re.compile('(<td[^>]*>)(.*?)</td>')
tabletid_re = re.compile('<a[^>]*>(.*?)</a>')
datatext_re = re.compile(r' data-text="(\d+)"')


def strip_small(x):
    return x[7:-8] if x.startswith('<small>') and x.endswith('</small>') else x


def parse_vdisk_storage_from_http_api(node_id, pdisk_id, vslot_id):
    page = 'vdisk/json/blobindexstat'
    data = common.fetch(page, dict(node_id=node_id, pdisk_id=pdisk_id, vslot_id=vslot_id), fmt='json')
    res = []
    if 'stat' not in data:
        raise Exception(f'Error resposne "{data}"')
    tablets = data['stat'].get('tablets', [])
    for tablet_info in tablets:
        tablet_id = int(tablet_info['tablet_id'])
        for channel, channel_info in enumerate(tablet_info['channels']):
            size = int(channel_info.get('data_size', 0))
            res.append((tablet_id, channel, size))
    return res


def parse_vdisk_storage_legacy(host, pdisk_id, vslot_id):
    page = 'actors/vdisks/vdisk%09u_%09u' % (pdisk_id, vslot_id)
    data = common.fetch(page, dict(type='stat', dbname='LogoBlobs'), host, fmt='raw')
    res = []
    m = table_re.search(str(data))
    for row in row_re.finditer(m.group(1)):
        data = [
            strip_small(m.group(2)) if m_datatext is None else int(m_datatext.group(1))
            for m in cell_re.finditer(row.group(1))
            for m_datatext in [datatext_re.search(m.group(1))]  # try to find data-text attr in <td>
        ]
        m = tabletid_re.match(data[0])
        tablet_id = int(m.group(1))
        channel = int(data[1])
        size = data[3]
        if not isinstance(size, int):
            factor = None
            if size.endswith('KiB'):
                size = size[:-3]
                factor = 1024.0
            elif size.endswith('MiB'):
                size = size[:-3]
                factor = 1024.0**2
            elif size.endswith('GiB'):
                size = size[:-3]
                factor = 1024.0**3
            elif size.endswith('TiB'):
                size = size[:-3]
                factor = 1024.0**4
            elif size.endswith('PiB'):
                size = size[:-3]
                factor = 1024.0**5
            elif size.endswith('B'):
                size = size[:-1]
                factor = 1.0
            size = int(float(size) * factor)
        res.append((tablet_id, channel, size))
    return res


def parse_vdisk_storage(host, node_id, pdisk_id, vslot_id):
    try:
        if common.connection_params.http:
            return parse_vdisk_storage_from_http_api(node_id, pdisk_id, vslot_id)
        else:
            return parse_vdisk_storage_legacy(host, pdisk_id, vslot_id)
    except Exception as e:
        print('Failed to parse VDisk storage at host %s PDiskId# %d VSlotId# %d error# %s' % (host, pdisk_id, vslot_id, e), file=sys.stderr)
        return None
